
/* Web Polygraph       http://www.web-polygraph.org/
 * (C) 2003-2006 The Measurement Factory
 * Licensed under the Apache License, Version 2.0 */

#include "base/polygraph.h"

#include "base/polyLogCats.h"
#include "base/StatIntvlRec.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"
#include "runtime/SharedOpts.h"
#include "client/Client.h"
#include "client/FtpCxm.h"
#include "client/FtpCltXact.h"
#include "client/ServerRep.h"
#include "csm/oid2Url.h"
#include "csm/ContentCfg.h"
#include "runtime/globals.h"
#include "runtime/polyErrors.h"
#include "runtime/FtpText.h"

FtpCltXact::FtpCltXact() {
	FtpCltXact::reset();
}

void FtpCltXact::reset() {
	CltXact::reset();
	protoStat = &StatIntvlRec::theFtpStat;
	theReqOid.type(TheBodilessContentId);
	theRep.reset();
	theReqCmd = FtpReq::frcUnknown;
	theDataConn.reset();
	theDataConn.logCat(lgcCltSide);
	theDataChannelState = dcsNone;
}

void FtpCltXact::exec(Client *anOwner, Connection *aConn) {
	CltXact::exec(anOwner, aConn);
	newState(stHdrWaiting);

	Assert(!theMgr);
	theMgr = FtpCxm::Get();
	theMgr->control(this);
}

void FtpCltXact::finish(Error err) {
//Comment << here << "finishing " << err << endc;
	stopDataChannel();
	stopCtrlChannel(err);
	CltXact::finish(err);
}

bool FtpCltXact::writeFirst() const {
	return false;
}

PipelinedCxm *FtpCltXact::getPipeline() {
	return 0; // FTP cannot be a pipelined transaction
}

void FtpCltXact::pipeline(PipelinedCxm *) {
    // FTP does not support pipelining so we will not change theMgr
}

void FtpCltXact::stopCtrlChannel(const Error &err) {
	if (theMgr) {
		if (err)
			theMgr->noteAbort(this);
		else
			theMgr->noteDone(this);
		theMgr->release(this);
		theMgr = 0;
	}
}

bool FtpCltXact::controlledPostRead(bool &needMore) {
	Assert(theState == stHdrWaiting);

	if (Error err = handleReplies()) {
		finish(err);
		return false;
	}

	if (finishIfDone() || !theMgr)
		return false;

	needMore = theState == stHdrWaiting;
	return true;
}

bool FtpCltXact::finishIfDone() {
//Comment << here << "ctrl: " << theState << " data: " << theDataChannelState << endc;
//Comment << here << (theState == stDone) << " / " << (theDataChannelState == dcsDone) << " || " << (theDataChannelState == dcsNone) << endc;
	// do not quit if we are not done with the control channel
	if (theState != stDone) 
		return false;

	// quit if we are done with the data channel
	if (theDataChannelState == dcsDone || theDataChannelState == dcsNone) {
		finish(0);
		return true;
	}

	return false;
}

Error FtpCltXact::handleReplies() {
	IOBuf &buf = theConn->theRdBuf;
//Comment << "FTP reply (" << buf.contSize() << "): " << endc; printMsg(buf.content(), buf.contSize());

	// due to Preliminary and Intermediate replies, we might read several 
	// server replies at once; handle them one by one until we are done or err
	while (theRep.parse(buf.content(), buf.contSize())) {
//Comment << "FTP reply size: " << theRep.size() << endc;
		theRepSize.expect(theRep.size()); // XXX: ignores data channel
		consume(theRep.size()); // XXX: does not work for data/ctrl conn split
		if (const Error err = interpretReply())
			return err;

		if (theRep.code() >= 200) {
			if (!buf.empty()) { // leftovers?
				if (ReportError(errExtraRepData)) { // XXX: add and use errExtraReply
					Comment << "FTP reply leftovers (" << buf.contSize() <<
						")" << endc;
					if (TheOpts.theDumpFlags(dumpErr, dumpAny))
						printMsg(buf.content(), buf.contSize());
				}
				return errOther;
			}

			theReqSize.reset(); // XXX: why? should we do it after 150?
			if (theState == stHdrWaiting)
				newState(stSpaceWaiting); // signal that no more data is needed
			return 0;
		}

		theRep.reset();
		theRepSize.reset();
		// if we got a Positive Preliminary reply, check if is there more
	}

	if (buf.full()) // command too big
		return errFtpHugeCmd;

	if (theConn->atEof()) { // premature end of reply stream
		if (buf.empty())
			return errFtpPrematureEndOfCtrl; // expected a reply, got nothing
		else
			return errFtpPrematureEndOfMsg; // we did not get the entire reply
	}

	Assert(theState == stHdrWaiting);
	return 0; // need more data to parse the header
}

void FtpCltXact::makeReq(WrBuf &buf) {
	Assert(!theReqSize.expected().known());
	ofixedstream os(buf.space(), buf.spaceSize());
	switch (theReqCmd) {
		case FtpReq::frcUSER:
			makeUser(os);
			break;
		case FtpReq::frcPASS:
			makePass(os);
			break;
		case FtpReq::frcTYPE:
			makeType(os);
			break;
		case FtpReq::frcPASV:
			makePasv(os);
			break;
		case FtpReq::frcRETR:
			makeRetr(os);
			break;
		case FtpReq::frcQUIT:
			makeQuit(os);
			break;
		default:
			Assert(false);
	}
	theReqSize.expect(Size(os.tellp()));
	buf.appended(theReqSize.expected());
}

Error FtpCltXact::interpretReply() {
	Assert(theState == stHdrWaiting);
	if (theRep.code() < 150 || theRep.code() >= 400)
		return errFtpCommandFailed;

	// Interpret the server response and make sure we are making progress 
	// according to client-side plan

	Error err = 0;
	// XXX: use named (enum) constants below
	// TODO: use a const transition table: rep.code -> expected state,
	//       next state?
	switch (theRep.code()) {
		case 220: // server greating
			err = transition(FtpReq::frcUnknown, FtpReq::frcUSER);
			if (!err && theSrvRep) {
				// we note request here to mimic HTTP; TODO: redesign?
				theSrvRep->noteRequest();
				if (true) // XXX: but there are FTP proxies; detect polysrv
					theSrvRep->noteFirstHandResponse();
			}
			break;
		case 331: // username accepted, password required
			err = transition(FtpReq::frcUSER, FtpReq::frcPASS);
			break;
		case 230: // password accepted.
			err = transition(FtpReq::frcPASS, FtpReq::frcTYPE);
			break;
		case 200: // entering binary mode.
			err = transition(FtpReq::frcTYPE, FtpReq::frcPASV);
			break;
		case 227: // passive address
			err = transition(FtpReq::frcPASV, FtpReq::frcRETR);
			if (!err)
				err = interpretPasv();
			break;
		case 150: // Opening data connection.
			err = transition(FtpReq::frcRETR, FtpReq::frcUnknown);
			break;
		case 226: // Operation completed
			err = transition(FtpReq::frcUnknown, FtpReq::frcQUIT);
			break;
		case 221: // Bye.
			err = transition(FtpReq::frcQUIT, FtpReq::frcUnknown);
			theConn->lastUse(true); // no pconn support yet
			stopCtrlChannel(err);
			if (!err)
				newState(stDone); // XXX: misplaced, not the last call
			break;
		default:
			Assert(false); // XXX: report unknown status codes
	}

	return err;
}

Error FtpCltXact::transition(FtpReq::Command from, FtpReq::Command to) {
//Comment << here << "transition: " << theReqCmd << " =? " << from << " -> " << to << endc;
	if (theReqCmd != from)
		return errFtpCommandFailed; // XXX: report details and use errOther

	theReqCmd = to;
	return 0;		
}

void FtpCltXact::makeUser(ostream &os) {
	os << ftpReqUser;
	theReqCmd = FtpReq::frcUSER;
}

void FtpCltXact::makePass(ostream &os) {
	os << ftpReqPass;
	theReqCmd = FtpReq::frcPASS;
}

void FtpCltXact::makeType(ostream &os) {
	os << ftpReqTypeBinary;
	theReqCmd = FtpReq::frcTYPE;
}

void FtpCltXact::makePasv(ostream &os) {
	os << ftpReqPasv;
	theReqCmd = FtpReq::frcPASV;
}

void FtpCltXact::makeRetr(ostream &os) {
	Assert(theDataChannelState == dcsConnecting ||
		theDataChannelState == dcsConnected);
	os << ftpReqRetr;
	Oid2UrlPath(theOid, os) << crlf;
	theReqCmd = FtpReq::frcRETR;
	theDataChannelState = dcsInProgress;
	theDataConn.theRd.start(this);

	if (theOid.foreignUrl())
		TheEmbedStats.foreignUrlRequested++;
}

void FtpCltXact::makeQuit(ostream &os) {
	os << ftpReqQuit;
	theReqCmd = FtpReq::frcQUIT;
}

Error FtpCltXact::interpretPasv() {
	struct in_addr ip_addr;
	if (!theRep.data() ||
		theRep.data()[0] != '=')
		return errFtpBadPasv;
	const char *s = theRep.data().cstr() + 1;
	const char *p;
	for (int i = 0; i < 4; ++i) {
		int n;
		if (!isInt(s, n, &p) ||
			p + 1 >= theRep.data().cstr() + theRep.data().len() ||
			*p != ',')
			return errFtpBadPasv;
		(reinterpret_cast< unsigned char *>(&ip_addr))[i] = n;
		s = p + 1;
	}

	int n1, n2;
	if (!isInt(s, n1, &p) ||
		p + 1 >= theRep.data().cstr() + theRep.data().len() ||
		*p != ',')
		return errFtpBadPasv;
	s = p + 1;
	if (!isInt(s, n2))
		return errFtpBadPasv;

	startDataChannel(NetAddr(InAddress(ip_addr), n1 * 256 + n2));
	return 0;
}

void FtpCltXact::startDataChannel(const NetAddr &addr) {
	stopDataChannel();
	theDataConn.connect(addr, SockOpt(), theOwner->portMgr());
	theDataConn.theWr.start(this);
	theDataChannelState = dcsConnecting;
}

void FtpCltXact::stopDataChannel() {
	theDataConn.closeNow();
	theDataChannelState = dcsNone;
}

void FtpCltXact::noteReadReady(int fd) {
	Assert(fd == theDataConn.fd());
	Assert(theDataChannelState == dcsInProgress);

	// TODO: support simulated I/O aborts, MD5 checks
	const Size sz = theDataConn.read(); 
//Comment << "FTP data read: " << sz << " eof: " << theDataConn.atEof() << endc;
	if (theDataConn.bad()) {
		finish(errOther);
		return;
	}

	if (sz > 0) {
		theRepSize.got(sz);
		theDataConn.theRdBuf.consumed(sz);
		return; // only EOF signals the end of data
	}

	if (theDataConn.atEof()) {
		stopDataChannel();
		theDataChannelState = dcsDone;
		finishIfDone();
		return;
	}
}

void FtpCltXact::noteWriteReady(int fd) {
	Assert(fd == theDataConn.fd());

	Assert(theDataChannelState == dcsConnecting);
	// TODO: check for errors using getsockopt
	theDataConn.theWr.stop(this);
	theDataChannelState = dcsConnected;
	// makeRetr will start reading on a data channel
}

bool FtpCltXact::controlledFill(bool &needMore) {
	// do nothing while data channel is connecting
//	if (theDataChannelState == dcsConnecting)
//		return true;

	return CltXact::controlledFill(needMore);
}

bool FtpCltXact::controlledPostWrite(Size &size, bool &needMore) {
	// do nothing while data channel is connecting
//	if (theDataChannelState == dcsConnecting)
//		return true;

	return CltXact::controlledPostWrite(size, needMore);
}

bool FtpCltXact::controlledMasterRead() {
	// do nothing while data channel is reading
//	if (theDataChannelState == dcsInProgress)
//		return true;

	return CltXact::controlledMasterRead();
}
