
/* Web Polygraph       http://www.web-polygraph.org/
 * (C) 2003-2006 The Measurement Factory
 * Licensed under the Apache License, Version 2.0 */

#include "base/polygraph.h"

#include "base/StatIntvlRec.h"
#include "base/polyLogCats.h"
#include "csm/BodyIter.h"
#include "csm/ContentMgr.h"
#include "runtime/globals.h"
#include "runtime/ErrorMgr.h"
#include "runtime/polyErrors.h"
#include "runtime/FtpText.h"
#include "runtime/SharedOpts.h"
#include "server/Server.h"
#include "server/SrvOpts.h"

#include "server/FtpSrvXact.h"

FtpSrvXact::FtpSrvXact() {
	FtpSrvXact::reset();
}

void FtpSrvXact::reset() {
	SrvXact::reset();
	protoStat = &StatIntvlRec::theFtpStat;
	theReqOid.type(TheBodilessContentId);
	theReq.reset();
	theDir = String();
	theFilename = String();
	stopListenDataChannel();
	stopDataChannel();
	theDataConn.reset();
	theDataConn.logCat(lgcSrvSide);
	theWrittenSize.reset();
}

void FtpSrvXact::finish(Error err) {
	stopListenDataChannel();
	stopDataChannel();
	theRepSize = theWrittenSize; // XXX: ugly; what is reply?
	SrvXact::finish(err);
}

void FtpSrvXact::noteReadReady(int fd) {
	if (fd == theSock.fd())
		acceptDataConnection();
	else
		SrvXact::noteReadReady(fd);
}

void FtpSrvXact::noteWriteReady(int fd) {
	if (fd == theDataConn.fd())
		noteDataChannelWriteReady();
	else
		SrvXact::noteWriteReady(fd);
}

void FtpSrvXact::doStart() {
	newState(stHdrWaiting);
	theWrittenSize.expect(0); // will increase later
	SrvXact::doStart();
}

void FtpSrvXact::noteConnReady() {
	// Connection has just been established, send greeting
	newState(stSpaceWaiting);
	noteWriteReady(theConn->fd());
}

void FtpSrvXact::noteHdrDataReady() {
	if (theReq.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		theReqSize.expect(theReq.size());
		if (const Error err = interpretCmd()) {
			finish(err);
			return;
		}
		// XXX: check for another command and quit, if any. Since we do not
		// and should not format replies before we can write them, we cannot
		// handle pipelined commands. The I/O code will not call us if there
		// are buffered leftovers but no new I/O activity.
		newState(stSpaceWaiting);
		theConn->theRd.stop(this);
		consume(theReq.size());
		return; // switch to writing
	}

	if (theConn->theRdBuf.full()) { // command too big
		finish(errFtpHugeCmd);
		return;
	}

	if (theConn->atEof()) {
		if (!theConn->theRdBuf.empty())
			finish(errFtpPrematureEndOfMsg); // we did not get the entire cmd
		else
        if (theDataChannelState == dcsInProgress)
			finish(errFtpPrematureEndOfCtrl); // control channel died first
		else
			finish(0); // assume the user decided not to send QUIT
		return;
	}

	// continue reading
}

void FtpSrvXact::noteRepSent() {
	theWrittenSize.got(theRepSize.actual());
	theRepSize.reset();

	if (theReq.cmd() == FtpReq::frcQUIT) {
		theConn->lastUse(true);
		finish(0);
		return;
	}

	// expect more commands
	newState(stHdrWaiting);
	if (!theConn->theRd.theReserv) 
		theConn->theRd.start(this);

	// start sending data if needed
	if (theDataChannelState == dcsInProgress) {
		if (!theDataConn.theWr.theReserv)
			theDataConn.theWr.start(this);
	}
}

void FtpSrvXact::makeRep(WrBuf &buf) {
	Assert(!theRepSize.expected().known());
	ofixedstream os(buf.space(), buf.spaceSize());
	switch(theReq.cmd()) {
		case FtpReq::frcUnknown:
			makeSimpleResponse(os, ftpRspGreeting);
			break;
		case FtpReq::frcSYST:
			makeSimpleResponse(os, ftpRspSystem);
			break;
		case FtpReq::frcUSER:
			makeSimpleResponse(os, ftpRspUser);
			break;
		case FtpReq::frcPASS:
			makeSimpleResponse(os, ftpRspPass);
			break;
		case FtpReq::frcTYPE:
			makeSimpleResponse(os, ftpRspTypeBinary);
			break;
		case FtpReq::frcCWD:
			makeSimpleResponse(os, ftpRspCwd);
			break;
		case FtpReq::frcMDTM:
			makeMdtmResponse(os);
			break;
		case FtpReq::frcSIZE:
			makeSizeResponse(os);
			break;
		case FtpReq::frcPASV:
			makePasvResponse(os);
			break;
		case FtpReq::frcRETR:
			makeRetrResponse(os);
			break;
		case FtpReq::frcQUIT:
			makeSimpleResponse(os, ftpRspQuit);
			break;
		case FtpReq::frcEPSV:
		case FtpReq::frcOther:
			makeSimpleResponse(os, ftpRspCmdNotImplemented);
			break;
		default:
			Assert(false);
	}
	theRepSize.expect(Size(os.tellp()));
	theWrittenSize.expectMore(theRepSize.expected());
	buf.appended(theRepSize.expected());
}

Error FtpSrvXact::interpretCmd() {
	Error err = 0;
	switch(theReq.cmd()) {
		case FtpReq::frcUSER:
		case FtpReq::frcPASS:
		case FtpReq::frcQUIT:
		case FtpReq::frcSYST:
			break;
		case FtpReq::frcTYPE:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
			if (theReq.argument() != "I" &&
				theReq.argument() != "L 8")
				err = errFtpModeNotBinary;
			break;
		case FtpReq::frcCWD:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
			if (theReq.argument() == "/")
				theDir = ""; // root directory
			else
			if (theReq.argument().startsWith("/"))
				theDir = theReq.argument(); // absolute directory
			else
				theDir += "/" + theReq.argument(); // relative
			break;
		case FtpReq::frcMDTM:
		case FtpReq::frcSIZE:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
				theFilename = theReq.argument();
			break;
		case FtpReq::frcPASV:
			err = startListenDataChannel();
			break;
		case FtpReq::frcRETR:
			err = processRetr();
			break;
		case FtpReq::frcEPSV:
			// we do not support it, but do not want to report it
			// TODO: add something like ReportErrorOnce(err, key)
			break;
		case FtpReq::frcOther:
			// report but continue
			if (ReportError(errFtpCommandNotImplemented) &&
				TheOpts.theDumpFlags(dumpErr, dumpAny))
                printMsg(theReq.image().data(), theReq.image().len());
			break;
		default:
			err = errFtpSyntax;
	}
	return err;
}

Error FtpSrvXact::startListenDataChannel() {
	Assert(theDataPort < 0);

	// TODO: support port manager so that we can use non-ephemeral ports
	const NetAddr addr(theOwner->host().addrN(), 0);
	theSock = theOwner->makeListenSocket(addr);
	if (!theSock)
		return errOther;

	theDataPort = theSock.lport();
	Assert(theDataPort >= 0);

	theReserv = TheFileScanner->setFD(theSock.fd(), dirRead, this);
	return 0;
}

void FtpSrvXact::stopListenDataChannel() {
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
	if (theSock)
		theSock.close();
	theDataPort = -1;
}

void FtpSrvXact::stopDataChannel() {
	theDataConn.closeNow();
	theDataChannelState = dcsNone;
}

void FtpSrvXact::acceptDataConnection() {
	Assert(!theDataConn.sock());
	theDataConn.protoStat = protoStat;
	bool fatal = false;
	Must(theDataConn.accept(theSock, SockOpt(), fatal));
	stopListenDataChannel();
}

void FtpSrvXact::noteDataChannelWriteReady() {
	Assert(theOid);
	Assert(theBodyIter);
	theBodyIter->pour();
	const Size sz = theDataConn.write();

	if (sz > 0)
		theWrittenSize.got(sz);

	if ((theBodyIter->pouredAll() && theDataConn.theWrBuf.empty()) ||
		theDataConn.bad()) {
		theDataChannelState = theDataConn.bad() ? dcsAborted : dcsDone;
		theDataConn.closeNow();
		theBodyIter->putBack();
		theBodyIter = 0;
		// tell the client we are done data writing
		theConn->theWr.start(this);
		newState(stSpaceWaiting);
	} else
	if (!theDataConn.theWr.theReserv)
		theDataConn.theWr.start(this);
}

Error FtpSrvXact::processRetr() {
	if (!theReq.argument())
		return errFtpNoArgument;
	theFilename = theReq.argument();

	if (theDataChannelState != dcsNone)
		return errFtpCmdSequence;

	stopListenDataChannel();

	if (!theDataConn.sock())
		return errFtpCmdSequence;

	if (prepareContent()) {
		theBodyIter->start(&theDataConn.theWrBuf);
		theDataChannelState = dcsInProgress;
	} else {
		// leave theDataChannelState == dcsNone to report ftpRspBadUrl later
	}

	return 0;
}

bool FtpSrvXact::prepareContent() {
	if (theBodyIter)
		return true;

	const String url(theDir + '/' + theFilename);
	const char *p = url.data();
	theOid.parse(p, p + url.len());
	theOid.get(true);

	// XXX: pxy server cannot support ignoreUrls; see Server::hostIdx
	bool ignoreUrls = (TheSrvOpts.ignoreUrls && theOwner->hostIdx() >= 0);

	if (ignoreUrls)
		overwriteUrl();
	else
	if (theOid.foreignUrl())
		grokForeignUrl(url, ignoreUrls);

	if (theOid.foreignUrl())
		return false;
	else {
		Assert(theOid.type() >= TheContentMgr.normalContentStart());
		theRepContentCfg = TheContentMgr.get(theOid.type());
		theBodyIter = theRepContentCfg->getBodyIter(theOid);
		theWrittenSize.expectMore(theBodyIter->contentSize());
		return true;
	}
}

void FtpSrvXact::makeSimpleResponse(ostream &os, const String &reply) {
	os << reply;
}

void FtpSrvXact::makeMdtmResponse(ostream &os) {
	// XXX: hard coded modification time
	// should be similar to IMS
	os
		<< ftpRspMdtm
		<< "20080714151541"
		<< crlf
		;
}

void FtpSrvXact::makeSizeResponse(ostream &os) {
	if (prepareContent())
		os << ftpRspSize << (int)theBodyIter->contentSize() << crlf;
	else
		os << ftpRspBadUrl;
}

void FtpSrvXact::makePasvResponse(ostream &os) {
	Assert(theDataPort >= 0);
	os << ftpRspPasv;
	// XXX: IPv4 only
	for (int i = 0; i < 4; ++i)
		os << theOwner->host().octet(i) << ',';
	const int p1 = theDataPort / 256;
	const int p2 = theDataPort - 256 * p1;
	os << p1 << ',' << p2 << crlf;
	theDataPort = -1;
}

void FtpSrvXact::makeRetrResponse(ostream &os) {
	switch (theDataChannelState) {
		case dcsNoConnection:
			os << ftpRspRetrNoDataConn;
			break;
		case dcsInProgress:
			os << ftpRspRetrInProgress;
			break;
		case dcsAborted:
			os << ftpRspRetrAborted;
			theDataChannelState = dcsNone;
			break;
		case dcsDone:
			os << ftpRspRetrComplete;
			theDataChannelState = dcsNone;
			break;
		case dcsNone:
			Assert(theOid.foreignUrl());
			os << ftpRspBadUrl;
	}
}
